#!/usr/bin/env python
import json
import logging
import os
import subprocess
import sys
import tempfile
import time
from contextlib import contextmanager

from cloudfoundry.utils import wait_for
import jujuclient
import requests
import yaml

client = None


def command(*base_args, **kwargs):
    def callable_command(*args):
        check = kwargs.get('check', False)
        all_args = base_args + args
        logging.debug(' '.join(all_args))
        if check is True:
            return subprocess.check_call(all_args)
        else:
            return subprocess.check_output(all_args).strip()
    return callable_command


class Commander(object):
    def __getattr__(self, key):
        return command(key)

    def check(self, *args):
        return command(*args, check=True)

    def __call__(self, cmd):
        subprocess.check_output(cmd, shell=True)


sh = Commander()
dig = command('dig', '+short')


@contextmanager
def cd(directory):
    cwd = os.getcwd()
    os.chdir(directory)
    try:
        yield
    finally:
        os.chdir(cwd)


def current_env():
    return sh.juju('switch')


def validate_app(endpoint, url='/', status_code=200, expected=None):
    url = endpoint + url
    r = requests.get(url)
    assert r.status_code == status_code, "Invalid HTTP status code"
    if expected:
        assert r.text == expected, "Invalid result, got %s, expected %s" % (
            r.text, expected
        )


def get_jenv():
    cenv = current_env()
    return os.path.join(os.environ.get('JUJU_HOME',
                                       os.path.expanduser('~/.juju')),
                        'environments/%s.jenv' % cenv)


def bootstrap():
    if not os.path.exists(get_jenv()):
        sh.juju('bootstrap')


def get_admin_password():
    jenv = get_jenv()
    data = yaml.load(open(jenv))
    return data['password']


def deploy(**config):
    status = get_client().status()
    if 'cloudfoundry' not in status['Services']:
        # create an up to date config
        defaults = {
            'cloudfoundry': {
                'admin_secret': get_admin_password(),
                'cf_version': 'latest',
                'placement': 'dense'
            }
        }
        if config:
            defaults.update(config)

        fd, fn = tempfile.mkstemp(suffix='.yaml')
        os.close(fd)
        with open(fn, 'w') as fp:
            yaml.dump(defaults, fp)
        sh.juju('deploy', '--config=%s' % fn,
                '--repository=%s' % os.getcwd(),
                '--constraints', 'instance-type=m3.medium',
                'local:trusty/cloudfoundry')
        time.sleep(5)
        os.unlink(fn)


def cf_service():
    status = get_client().status()
    if 'cloudfoundry' not in status['Services']:
        return False
    sdata = status['Services']['cloudfoundry']
    for unit in sdata['Units'].values():
        if unit['Agent']['Status'] not in ['installed', 'started']:
            return False
    return True


def endpoint():
    status = get_client().status()
    try:
        units = status['Services']['haproxy']['Units']
        unit = sorted(units.keys())[0]
        hostname = units[unit]['PublicAddress']
        if not hostname:
            return False
        return dig(hostname)
    except KeyError:
        return False
    return endpoint


def api():
    return "http://api.%s.xip.io" % endpoint()


def login():
    try:
        cf = sh.check('cf')
        cf('api', api())
        cf('auth', 'admin', 'admin')
        cf('create-space', '-o', 'juju-org', 'my-space')
        cf('target', '-s', 'my-space', '-o', 'juju-org')
    except subprocess.CalledProcessError:
        return False
    return True


def get_client():
    global client
    if client is None:
        def env_connection():
            # It seems that calling status forces the rewrite of
            # .jenv with the state server
            sh.juju('status')
            data = yaml.load(open(get_jenv()))
            print "Waiting for state server(s)", data['state-servers']
            return data.get('state-servers', []) != []
        current = current_env()
        wait_for(600, 20, env_connection)
        client = jujuclient.Environment.connect(current)
    return client


def get_cats():
    logging.info("Getting CATs from github")
    if not os.path.exists('cf-acceptance-tests'):
        sh.git('clone', 'https://github.com/cloudfoundry/cf-acceptance-tests.git')
    with cd('cf-acceptance-tests'):
        sh.git('pull')
        sh('./bin/compile')


def run_cats():
    get_cats()
    ep = endpoint()
    with cd('cf-acceptance-tests'):
        fd, fn = tempfile.mkstemp(suffix=".json")
        os.close(fd)
        with open(fn, 'w') as fp:
            defaults = {
                "suite_name": "smoke",
                "apps_domain": "%s.xip.io" % ep,
                "api": api(),
                "syslog_drain_port": 8082,
                "syslog_ip_address": ep,
                "password": "admin",
                "admin_password": "admin",
                "user": "admin",
                "admin_user": "admin",
                "org": "juju-org",
                "space": "my-space",
                "use_existing_org": True,
                "use_existing_space": True,
                "skip_ssl_validation": True,
            }
            json.dump(defaults, fp, indent=2, sort_keys=True)
        logging.info("Running CATs %s", defaults)
        env = os.environ.copy()
        env.update({'CONFIG': os.path.abspath(fn)})
        subprocess.call('./bin/test', shell=True, env=env)


def main():
    logging.basicConfig(level=logging.INFO)
    bootstrap()
    deploy()
    wait_for(60 * 40, 30, cf_service, endpoint, login)
    run_cats()
    sys.exit(0)


if __name__ == '__main__':
    main()
